# Generated by Django 3.2.13 on 2022-08-08 07:47

from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('tenancy', '0002_auto_slug'),
        ('nautobot_cable_utils', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='CablePlug',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False, unique=True)),
                ('name', models.CharField(max_length=50, unique=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='cabletemplate',
            name='tenant',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='cable_templates', to='tenancy.tenant'),
        ),
        migrations.AddField(
            model_name='cabletemplate',
            name='plug',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='cable_templates', to='nautobot_cable_utils.cableplug'),
        ),
    ]



from django import forms
from django.forms import Form, ChoiceField
from django_filters.fields import ModelMultipleChoiceField

from nautobot.extras.forms import NautobotFilterForm
from nautobot.circuits.models import Circuit, CircuitTermination, Provider
from nautobot.dcim.choices import CableStatusChoices, CableTypeChoices
from nautobot.dcim.models import (
    Cable, ConsolePort, ConsoleServerPort, Device, FrontPort, Interface,
    PowerFeed, PowerOutlet, PowerPanel, PowerPort, Rack, RearPort,
    Site,
)

from nautobot.extras.models import Status
from nautobot.utilities.choices import ChoiceSet

from nautobot.utilities.forms import (
    APISelect, BootstrapMixin, ColorSelect, CSVModelForm,
    DynamicModelChoiceField, StaticSelect2, add_blank_choice, ExpandableNameField, parse_alphanumeric_range
)
from nautobot.tenancy.models import Tenant
from nautobot.utilities.forms import CSVModelChoiceField
from .cable_router import CableRouter

from .models import CableTemplate, MeasurementLog, CablePlug
from django.core.exceptions import ObjectDoesNotExist
import re

NUMERIC_EXPANSION_PATTERN = r"\[((?:\d+[?:,-])+\d+)\]"


def expand_alphanumeric_pattern(string):
    """
    Expand an alphabetic pattern into a list of strings.
    """
    lead, pattern, remnant = re.split(NUMERIC_EXPANSION_PATTERN, string, maxsplit=1)
    parsed_range = parse_alphanumeric_range(pattern)

    maximum_length = max([len(str(x)) for x in parsed_range])

    for i in parsed_range:
        padded_i = str(i).zfill(maximum_length)
        if re.search(NUMERIC_EXPANSION_PATTERN, remnant):
            for string2 in expand_alphanumeric_pattern(remnant):
                yield f"{lead}{padded_i}{string2}"
        else:
            yield f"{lead}{padded_i}{remnant}"


class ExpandableCableNumberPatternField(forms.CharField):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.help_text:
            self.help_text = """
                Numeric ranges are supported for bulk creation. Mixed cases and types within a single range
                are not supported. Leading zeros will be added. Examples:
                <ul>
                    <li><code>e[0-3][0-13,17]</code></li>
                </ul>
                """

    def to_python(self, value):
        if not value:
            return ""
        if re.search(NUMERIC_EXPANSION_PATTERN, value):
            return list(expand_alphanumeric_pattern(value))
        return [value]


class ConnectCableToDeviceForm(BootstrapMixin, forms.ModelForm):
    """
    Base form for connecting a Cable to a Device component
    """
    termination_a_site = DynamicModelChoiceField(
        queryset=Site.objects.all(),
        label='Site',
        required=False,
    )
    termination_a_rack = DynamicModelChoiceField(
        queryset=Rack.objects.all(),
        label='Rack',
        required=False,
        display_field='display_name',
        null_option='None',
        query_params={
            'site_id': '$termination_a_site'
        }
    )
    termination_a_device = DynamicModelChoiceField(
        queryset=Device.objects.all(),
        label='Device',
        required=False,
        display_field='display_name',
        query_params={
            'site_id': '$termination_a_site',
            'rack_id': '$termination_a_rack',
        }
    )
    termination_b_site = DynamicModelChoiceField(
        queryset=Site.objects.all(),
        label='Site',
        required=False
    )
    termination_b_rack = DynamicModelChoiceField(
        queryset=Rack.objects.all(),
        label='Rack',
        required=False,
        display_field='display_name',
        null_option='None',
        query_params={
            'site_id': '$termination_b_site'
        }
    )
    termination_b_device = DynamicModelChoiceField(
        queryset=Device.objects.all(),
        label='Device',
        required=False,
        display_field='display_name',
        query_params={
            'site_id': '$termination_b_site',
            'rack_id': '$termination_b_rack',
        }
    )

    class Meta:
        model = Cable
        fields = [
            'termination_a_site', 'termination_a_rack', 'termination_a_device', 'termination_a_id',
            'termination_b_site', 'termination_b_rack', 'termination_b_device', 'termination_b_id'
        ]

    def clean_termination_a_id(self):
        # Return the PK rather than the object
        return getattr(self.cleaned_data['termination_a_id'], 'pk', None)

    def clean_termination_b_id(self):
        # Return the PK rather than the object
        return getattr(self.cleaned_data['termination_b_id'], 'pk', None)


class ConnectConsolePortToConsoleServerPortForm(ConnectCableToDeviceForm):
    termination_a_id = DynamicModelChoiceField(
        queryset=ConsolePort.objects.all(),
        label='Name',
        disabled_indicator='cable',
        query_params={
            'device_id': '$termination_a_device'
        }
    )
    termination_b_id = DynamicModelChoiceField(
        queryset=ConsoleServerPort.objects.all(),
        label='Name',
        disabled_indicator='cable',
        query_params={
            'device_id': '$termination_b_device'
        }
    )


class ConnectConsolePortToFrontPortForm(ConnectCableToDeviceForm):
    termination_a_id = DynamicModelChoiceField(
        queryset=ConsolePort.objects.all(),
        label='Name',
        disabled_indicator='cable',
        query_params={
            'device_id': '$termination_a_device'
        }
    )
    termination_b_id = DynamicModelChoiceField(
        queryset=FrontPort.objects.all(),
        label='Name',
        disabled_indicator='cable',
        query_params={
            'device_id': '$termination_b_device'
        }
    )


class ConnectConsolePortToRearPortForm(ConnectCableToDeviceForm):
    termination_a_id = DynamicModelChoiceField(
        queryset=ConsolePort.objects.all(),
        label='Name',
        disabled_indicator='cable',
        query_params={
            'device_id': '$termination_a_device'
        }
    )
    termination_b_id = DynamicModelChoiceField(
        queryset=RearPort.objects.all(),
        label='Name',
        disabled_indicator='cable',
        query_params={
            'device_id': '$termination_b_device'
        }
    )


class ConnectConsoleServerPortToFrontPortForm(ConnectCableToDeviceForm):
    termination_a_id = DynamicModelChoiceField(
        queryset=ConsoleServerPort.objects.all(),
        label='Name',
        disabled_indicator='cable',
        query_params={
            'device_id': '$termination_a_device'
        }
    )
    termination_b_id = DynamicModelChoiceField(
        queryset=FrontPort.objects.all(),
        label='Name',
        disabled_indicator='cable',
        query_params={
            'device_id': '$termination_b_device'
        }
    )


class ConnectConsoleServerPortToRearPortForm(ConnectCableToDeviceForm):
    termination_a_id = DynamicModelChoiceField(
        queryset=ConsoleServerPort.objects.all(),
        label='Name',
        disabled_indicator='cable',
        query_params={
            'device_id': '$termination_a_device'
        }
    )
    termination_b_id = DynamicModelChoiceField(
        queryset=RearPort.objects.all(),
        label='Name',
        disabled_indicator='cable',
        query_params={
            'device_id': '$termination_b_device'
        }
    )


class ConnectPowerfeedToPowerPortForm(ConnectCableToDeviceForm):
    termination_a_site = DynamicModelChoiceField(
        queryset=Site.objects.all(),
        label='Site',
        required=False,
        display_field='cid'
    )
    termination_a_powerpanel = DynamicModelChoiceField(
        queryset=PowerPanel.objects.all(),
        label='Power Panel',
        required=False,
        query_params={
            'site_id': '$termination_a_site',
        }
    )
    termination_a_id = DynamicModelChoiceField(
        queryset=PowerFeed.objects.all(),
        label='Name',
        disabled_indicator='cable',
        query_params={
            'power_panel_id': '$termination_a_powerpanel'
        }
    )
    termination_b_id = DynamicModelChoiceField(
        queryset=PowerPort.objects.all(),
        label='Name',
        disabled_indicator='cable',
        query_params={
            'device_id': '$termination_b_device'
        }
    )

    class Meta:
        model = Cable
        fields = [
            'termination_a_site', 'termination_a_powerpanel', 'termination_a_id',
            'termination_b_site', 'termination_b_rack', 'termination_b_device', 'termination_b_id'
        ]


class ConnectPowerOutletToPowerPortForm(ConnectCableToDeviceForm):
    termination_a_id = DynamicModelChoiceField(
        queryset=PowerOutlet.objects.all(),
        label='Name',
        disabled_indicator='cable',
        query_params={
            'device_id': '$termination_a_device'
        }
    )
    termination_b_id = DynamicModelChoiceField(
        queryset=PowerPort.objects.all(),
        label='Name',
        disabled_indicator='cable',
        query_params={
            'device_id': '$termination_b_device'
        }
    )


class ConnectInterfaceForm(ConnectCableToDeviceForm):
    termination_a_id = DynamicModelChoiceField(
        queryset=Interface.objects.all(),
        label='Name',
        disabled_indicator='cable',
        query_params={
            'device_id': '$termination_a_device',
            'kind': 'physical',
        }
    )
    termination_b_id = DynamicModelChoiceField(
        queryset=Interface.objects.all(),
        label='Name',
        disabled_indicator='cable',
        query_params={
            'device_id': '$termination_b_device',
            'kind': 'physical',
        }
    )


class ConnectFrontPortToInterfaceForm(ConnectCableToDeviceForm):
    termination_a_id = DynamicModelChoiceField(
        queryset=FrontPort.objects.all(),
        label='Name',
        disabled_indicator='cable',
        query_params={
            'device_id': '$termination_a_device'
        }
    )
    termination_b_id = DynamicModelChoiceField(
        queryset=Interface.objects.all(),
        label='Name',
        disabled_indicator='cable',
        query_params={
            'device_id': '$termination_b_device',
            'kind': 'physical',
        }
    )


class ConnectInterfaceToRearPortForm(ConnectCableToDeviceForm):
    termination_a_id = DynamicModelChoiceField(
        queryset=Interface.objects.all(),
        label='Name',
        disabled_indicator='cable',
        query_params={
            'device_id': '$termination_a_device',
            'kind': 'physical',
        }
    )
    termination_b_id = DynamicModelChoiceField(
        queryset=RearPort.objects.all(),
        label='Name',
        disabled_indicator='cable',
        query_params={
            'device_id': '$termination_b_device'
        }
    )


class ConnectCircuitTerminationToInterfaceForm(ConnectCableToDeviceForm):
    termination_a_provider = DynamicModelChoiceField(
        queryset=Provider.objects.all(),
        label='Provider',
        required=False
    )
    termination_a_site = DynamicModelChoiceField(
        queryset=Site.objects.all(),
        label='Site',
        required=False
    )
    termination_a_circuit = DynamicModelChoiceField(
        queryset=Circuit.objects.all(),
        label='Circuit',
        display_field='cid',
        query_params={
            'provider_id': '$termination_a_provider',
            'site_id': '$termination_a_site',
        }
    )
    termination_a_id = DynamicModelChoiceField(
        queryset=CircuitTermination.objects.all(),
        label='Side',
        display_field='term_side',
        disabled_indicator='cable',
        query_params={
            'circuit_id': '$termination_a_circuit'
        }
    )
    termination_b_id = DynamicModelChoiceField(
        queryset=Interface.objects.all(),
        label='Name',
        disabled_indicator='cable',
        query_params={
            'device_id': '$termination_b_device',
            'kind': 'physical',
        }
    )

    class Meta:
        model = Cable
        fields = [
            'termination_a_provider', 'termination_a_site', 'termination_a_circuit', 'termination_a_id',
            'termination_b_site', 'termination_b_rack', 'termination_b_device', 'termination_b_id'
        ]


class ConnectCircuitTerminationForm(BootstrapMixin, forms.ModelForm):
    termination_a_provider = DynamicModelChoiceField(
        queryset=Provider.objects.all(),
        label='Provider',
        required=False
    )
    termination_a_site = DynamicModelChoiceField(
        queryset=Site.objects.all(),
        label='Site',
        required=False
    )
    termination_a_circuit = DynamicModelChoiceField(
        queryset=Circuit.objects.all(),
        label='Circuit',
        display_field='cid',
        query_params={
            'provider_id': '$termination_a_provider',
            'site_id': '$termination_a_site',
        }
    )
    termination_a_id = DynamicModelChoiceField(
        queryset=CircuitTermination.objects.all(),
        label='Side',
        display_field='term_side',
        disabled_indicator='cable',
        query_params={
            'circuit_id': '$termination_a_circuit'
        }
    )
    termination_b_provider = DynamicModelChoiceField(
        queryset=Provider.objects.all(),
        label='Provider',
        required=False
    )
    termination_b_site = DynamicModelChoiceField(
        queryset=Site.objects.all(),
        label='Site',
        required=False
    )
    termination_b_circuit = DynamicModelChoiceField(
        queryset=Circuit.objects.all(),
        label='Circuit',
        display_field='cid',
        query_params={
            'provider_id': '$termination_b_provider',
            'site_id': '$termination_b_site',
        }
    )
    termination_b_id = DynamicModelChoiceField(
        queryset=CircuitTermination.objects.all(),
        label='Side',
        display_field='term_side',
        disabled_indicator='cable',
        query_params={
            'circuit_id': '$termination_b_circuit'
        }
    )

    class Meta:
        model = Cable
        fields = [
            'termination_a_provider', 'termination_a_site', 'termination_a_circuit', 'termination_a_id',
            'termination_b_provider', 'termination_b_site', 'termination_b_circuit', 'termination_b_id', 'type',
            'status', 'label', 'color', 'length', 'length_unit',
        ]

    def clean_termination_a_id(self):
        # Return the PK rather than the object
        return getattr(self.cleaned_data['termination_a_id'], 'pk', None)

    def clean_termination_b_id(self):
        # Return the PK rather than the object
        return getattr(self.cleaned_data['termination_b_id'], 'pk', None)


class ConnectFrontPortForm(ConnectCableToDeviceForm):
    termination_a_id = DynamicModelChoiceField(
        queryset=FrontPort.objects.all(),
        label='Name',
        disabled_indicator='cable',
        query_params={
            'device_id': '$termination_a_device'
        }
    )
    termination_b_id = DynamicModelChoiceField(
        queryset=FrontPort.objects.all(),
        label='Name',
        disabled_indicator='cable',
        query_params={
            'device_id': '$termination_b_device'
        }
    )


class ConnectFrontPortToRearPortForm(ConnectCableToDeviceForm):
    termination_a_id = DynamicModelChoiceField(
        queryset=FrontPort.objects.all(),
        label='Name',
        disabled_indicator='cable',
        query_params={
            'device_id': '$termination_a_device'
        }
    )
    termination_b_id = DynamicModelChoiceField(
        queryset=RearPort.objects.all(),
        label='Name',
        disabled_indicator='cable',
        query_params={
            'device_id': '$termination_b_device'
        }
    )


class ConnectRearPortForm(ConnectCableToDeviceForm):
    termination_a_id = DynamicModelChoiceField(
        queryset=RearPort.objects.all(),
        label='Name',
        disabled_indicator='cable',
        query_params={
            'device_id': '$termination_a_device'
        }
    )
    termination_b_id = DynamicModelChoiceField(
        queryset=RearPort.objects.all(),
        label='Name',
        disabled_indicator='cable',
        query_params={
            'device_id': '$termination_b_device'
        }
    )


class ConnectCircuitTerminationToForm(BootstrapMixin, forms.ModelForm):
    termination_a_provider = DynamicModelChoiceField(
        queryset=Provider.objects.all(),
        label='Provider',
        required=False
    )
    termination_a_site = DynamicModelChoiceField(
        queryset=Site.objects.all(),
        label='Site',
        required=False
    )
    termination_a_circuit = DynamicModelChoiceField(
        queryset=Circuit.objects.all(),
        label='Circuit',
        display_field='cid',
        query_params={
            'provider_id': '$termination_a_provider',
            'site_id': '$termination_a_site',
        }
    )
    termination_a_id = DynamicModelChoiceField(
        queryset=CircuitTermination.objects.all(),
        label='Side',
        display_field='term_side',
        disabled_indicator='cable',
        query_params={
            'circuit_id': '$termination_a_circuit'
        }
    )
    termination_b_site = DynamicModelChoiceField(
        queryset=Site.objects.all(),
        label='Site',
        required=False
    )
    termination_b_rack = DynamicModelChoiceField(
        queryset=Rack.objects.all(),
        label='Rack',
        required=False,
        display_field='display_name',
        null_option='None',
        query_params={
            'site_id': '$termination_b_site'
        }
    )
    termination_b_device = DynamicModelChoiceField(
        queryset=Device.objects.all(),
        label='Device',
        required=False,
        display_field='display_name',
        query_params={
            'site_id': '$termination_b_site',
            'rack_id': '$termination_b_rack',
        }
    )

    class Meta:
        model = Cable
        fields = [
            'termination_a_provider', 'termination_a_site', 'termination_a_circuit', 'termination_a_id',
            'termination_b_site', 'termination_b_rack', 'termination_b_device', 'termination_b_id'
        ]

    def clean_termination_a_id(self):
        # Return the PK rather than the object
        return getattr(self.cleaned_data['termination_a_id'], 'pk', None)

    def clean_termination_b_id(self):
        # Return the PK rather than the object
        return getattr(self.cleaned_data['termination_b_id'], 'pk', None)


class ConnectCircuitTerminationToFrontPortForm(ConnectCircuitTerminationToForm):
    termination_b_id = DynamicModelChoiceField(
        queryset=FrontPort.objects.all(),
        label='Name',
        disabled_indicator='cable',
        query_params={
            'device_id': '$termination_b_device'
        }
    )


class ConnectCircuitTerminationToRearPortForm(ConnectCircuitTerminationToForm):
    termination_b_id = DynamicModelChoiceField(
        queryset=RearPort.objects.all(),
        label='Name',
        disabled_indicator='cable',
        query_params={
            'device_id': '$termination_b_device'
        }
    )


class CableTemplateForm(BootstrapMixin, forms.ModelForm):
    name = forms.CharField(label="Cable Number")

    class Meta:
        model = CableTemplate
        fields = (
            'name', 'label', 'owner', 'type', 'plug', 'color', 'length', 'length_unit'
        )


class CableTemplateBulkCreateForm(BootstrapMixin, forms.Form):
    pattern = ExpandableCableNumberPatternField(label="Cable Number Pattern")


class CableTemplateBulkForm(CableTemplateForm):

    def save(self, commit=True):
        if not self.instance.label:
            self.instance.label = self.instance.name

        return super().save(commit)


class CableTemplateFilterForm(NautobotFilterForm):
    model = CableTemplate
    q = forms.CharField(
        required=False,
        label='Search'
    )
    name = forms.CharField(required=False, label="Cable Number")
    type = forms.MultipleChoiceField(
        choices=add_blank_choice(CableTypeChoices),
        required=False,
        widget=StaticSelect2(),
    )
    color = forms.CharField(max_length=6, required=False, widget=ColorSelect())

    owner = ModelMultipleChoiceField(queryset=Tenant.objects.all(), to_field_name="slug", required=False,
                                     widget=StaticSelect2())
    plug = ModelMultipleChoiceField(queryset=CablePlug.objects.all(), to_field_name="name", required=False,
                                    widget=StaticSelect2())


class CommissionForm(BootstrapMixin, forms.Form):
    cable = forms.ModelChoiceField(
        queryset=Cable.objects.all(),
        required=False,
        disabled=True,
        widget=forms.HiddenInput()
    )
    template = forms.CharField(
        required=False,
    )

    def clean(self):
        super().clean()
        try:
            if CableTemplate.objects.get(name=self.cleaned_data['template']).cable is not None:
                raise forms.ValidationError({"template": "the assigned cable template is already in use"})
        except ObjectDoesNotExist:
            raise forms.ValidationError({"template": "the assigned cable template does not exist"})

    def save(self):

        template = CableTemplate.objects.get(name=self.cleaned_data['template'])
        cable = self.cleaned_data['cable']

        for field in ['type', 'label', 'color', 'length', 'length_unit']:
            value = getattr(template, field)
            if value:
                setattr(cable, field, value)

        if not template.label:
            cable.label = template.name

        connected_status = None
        statuses = Status.objects.get_for_model(Cable)

        for status in statuses:
            if status.slug == CableStatusChoices.STATUS_CONNECTED:
                connected_status = status

        cable.status = connected_status
        cable.save()
        template.cable = cable
        template.save()

        return cable


class UndoCommissionForm(BootstrapMixin, forms.Form):

    def save(self):
        cable = self.cleaned_data['cable']

        for field in ['type', 'label', 'color', 'length', 'length_unit']:
            value = getattr(template, field)
            if value:
                setattr(cable, field, value)

        connected_status = None
        statuses = Status.objects.get_for_model(Cable)

        for status in statuses:
            if status.slug == CableStatusChoices.STATUS_CONNECTED:
                connected_status = status

        cable.status = connected_status
        cable.save()

        return cable


class CableTemplateCSVForm(CSVModelForm):
    owner = CSVModelChoiceField(queryset=Tenant.objects.all(), required=False, to_field_name="name",
                                help_text="Assigned Owner")
    plug = CSVModelChoiceField(queryset=CablePlug.objects.all(), required=False, to_field_name="name", help_text="Plug")

    class Meta:
        model = CableTemplate
        fields = CableTemplate.csv_headers


class MeasurementLogForm(BootstrapMixin, forms.ModelForm):
    cable = DynamicModelChoiceField(
        queryset=Cable.objects.all(),
        display_field='label'
    )

    class Meta:
        model = MeasurementLog
        fields = ('link', 'cable')


class CableRouterMediaChoices(ChoiceSet):
    FIBER_SM = "fiber_sm"
    FIBER_MM = "fiber_mm"
    COPPER = "copper"
    CHOICES = (
        (FIBER_SM, "Single-Mode Fiber"),
        (FIBER_MM, "Multi-Mode Fiber"),
        (COPPER, "Copper"),
    )


class CableRouterLinkForm(BootstrapMixin, Form):
    """
    Base form for connecting a Cable to a Device component
    """

    termination_a_type = None
    termination_b_type = None
    termination_a = None

    def __init__(self, *args, **kwargs):
        self.termination_a = kwargs.pop("termination_a")
        self.termination_a_type = kwargs.pop("termination_a_type")
        self.termination_b_type = kwargs.pop("termination_b_type")
        super().__init__(*args, **kwargs)

    termination_b_site = DynamicModelChoiceField(
        queryset=Site.objects.all(),
        label='Site',
        required=False
    )
    termination_b_rack = DynamicModelChoiceField(
        queryset=Rack.objects.all(),
        label='Rack',
        required=False,
        display_field='display_name',
        null_option='None',
        query_params={
            'site_id': '$termination_b_site'
        }
    )
    termination_b_device = DynamicModelChoiceField(
        queryset=Device.objects.all(),
        label='Device',
        required=False,
        display_field='display_name',
        query_params={
            'site_id': '$termination_b_site',
            'rack_id': '$termination_b_rack',
        }
    )

    media_type = ChoiceField(choices=CableRouterMediaChoices, label="Media Type")

    class Meta:
        fields = [
            'termination_a_site', 'termination_a_rack', 'termination_a_device', 'termination_a_id',
            'termination_b_site', 'termination_b_rack', 'termination_b_device', 'termination_b_id'
        ]

    def clean(self):
        cleaned_data = super().clean()

        termination_b_model = self.termination_b_type.model_class()
        termination_b = termination_b_model.objects.get(id=cleaned_data['termination_b_id'])
        media_type = self.cleaned_data["media_type"]

        cleaned_data['cable_router'] = CableRouter(self.termination_a, self.termination_a_type, termination_b, self.termination_b_type, media_type)
        cleaned_data['path'] = cleaned_data['cable_router'].get_path()

    def clean_termination_a_id(self):
        # Return the PK rather than the object
        return getattr(self.cleaned_data['termination_a_id'], 'pk', None)

    def clean_termination_b_id(self):
        # Return the PK rather than the object
        return getattr(self.cleaned_data['termination_b_id'], 'pk', None)


class CableRouterLinkToInterfaceForm(CableRouterLinkForm):
    termination_b_model = Interface

    termination_b_id = DynamicModelChoiceField(
        queryset=Interface.objects.all(),
        label='Name',
        disabled_indicator='cable',
        query_params={
            'device_id': '$termination_b_device',
            'kind': 'physical',
        }
    )

    class Meta(CableRouterLinkForm.Meta):
        pass


class CableRouterLinkToFrontPortForm(CableRouterLinkForm):
    termination_b_model = FrontPort

    termination_b_id = DynamicModelChoiceField(
        queryset=FrontPort.objects.all(),
        label='Name',
        disabled_indicator='cable',
        query_params={
            'device_id': '$termination_b_device',
        }
    )

    class Meta(CableRouterLinkForm.Meta):
        pass


class CableRouterVerifyForm(BootstrapMixin, Form):
    class Meta:
        fields = []



import django_filters
from django.db.models import Q

from nautobot.extras.filters import NautobotFilterSet
from nautobot.utilities.choices import ColorChoices
from nautobot.utilities.filters import BaseFilterSet
from nautobot.tenancy.models import Tenant

from .models import CableTemplate, CablePlug


class CableTemplateFilterSet(NautobotFilterSet):
    q = django_filters.CharFilter(
        method='search',
        label='Search',
    )

    owner_id = django_filters.ModelMultipleChoiceFilter(
        queryset=Tenant.objects.all(),
        label="Owner (ID)",
        required=False,

    )
    owner = django_filters.ModelMultipleChoiceFilter(
        queryset=Tenant.objects.all(),
        field_name="owner__slug",
        to_field_name="slug",
        label="Owner (Slug)",
        required=False,

    )

    plug_id = django_filters.ModelMultipleChoiceFilter(
        queryset=CablePlug.objects.all(),
        label="Plug (ID)",
        required=False,
    )
    plug = django_filters.ModelMultipleChoiceFilter(
        queryset=CablePlug.objects.all(),
        field_name="plug__name",
        to_field_name="name",
        label="Plug (Name)",
        required=False,
    )

    class Meta:
        model = CableTemplate
        fields = ['name', 'type', 'color']

    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        qs_filter = (
                Q(name__icontains=value) |
                Q(label__icontains=value)
        )
        return queryset.filter(qs_filter)



# Generated by Django 3.2.15 on 2022-10-07 07:44

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('nautobot_cable_utils', '0004_rename_cable_number_cabletemplate_name'),
    ]

    operations = [
        migrations.AlterField(
            model_name='cabletemplate',
            name='name',
            field=models.CharField(max_length=100, unique=True),
        ),
    ]



# Generated by Django 3.1.13 on 2021-11-22 18:39

from django.db import migrations, models
import django.db.models.deletion
import nautobot.utilities.fields
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('dcim', '0005_device_local_context_schema'),
    ]

    operations = [
        migrations.CreateModel(
            name='MeasurementLog',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False, unique=True)),
                ('link', models.URLField(blank=True, null=True)),
                ('cable', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to='dcim.cable')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='CableTemplate',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False, unique=True)),
                ('cable_number', models.CharField(max_length=50, unique=True)),
                ('type', models.CharField(blank=True, max_length=50)),
                ('label', models.CharField(blank=True, max_length=100)),
                ('color', nautobot.utilities.fields.ColorField(blank=True, max_length=6)),
                ('length', models.PositiveSmallIntegerField(blank=True, null=True)),
                ('length_unit', models.CharField(blank=True, max_length=50)),
                ('cable', models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='dcim.cable')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



from uuid import UUID

from dijkstar import Graph, find_path, NoPathError
from django.contrib.contenttypes.models import ContentType
from django.forms import ValidationError
from django.db.models import Count, Q

from nautobot.dcim.choices import CableTypeChoices
from nautobot.dcim.models import Device, FrontPort, RearPort, Rack, Cable


class CablePath:
    def __init__(self, termination_a, termination_a_type, termination_b, termination_b_type, cable_inp):
        self.termination_a = termination_a
        self.termination_b = termination_b
        self.termination_a_type = termination_a_type
        self.termination_b_type = termination_b_type

        rack_list = [self.termination_a.device.rack]
        cable_list = []

        cables_to_pop = [*cable_inp]
        cable_objs = Cable.objects.filter(id__in=cable_inp).prefetch_related("_termination_a_device__rack",
                                                                             "_termination_b_device__rack")
        while len(cables_to_pop) > 0:

            next_cable = None
            is_swapped = False
            for c_id in cables_to_pop:
                c = next(filter(lambda co: co.id == UUID(c_id), cable_objs))
                if c.termination_a.device.rack == rack_list[-1]:
                    next_cable = c
                    is_swapped = False
                elif c.termination_b.device.rack == rack_list[-1]:
                    next_cable = c
                    is_swapped = True

            cable_list.append(next_cable)
            rack_list.append(
                next_cable.termination_a.device.rack if is_swapped else next_cable.termination_b.device.rack)
            cables_to_pop.pop(cables_to_pop.index(str(next_cable.id)))

        self.rack_list = rack_list
        self.cable_list = cable_list

    def get_renderable_path(self):
        needed_cables = self.get_needed_cables()

        # needed_cables and self.cable_list are sorted and only needs to be zipped for a complete cable trace.
        # Afterwards, we only need to add interfaces and devices for a neat trace.

        intermediate_elements = list()

        open_element = {
            "device": self.termination_a.device,
        }

        # We use UUIDs, so we do only compare for the same UUID, we do not care about the type, should fit anyways.

        cable_list = [*self.cable_list]

        if self.termination_a_type.app_label == "dcim" and self.termination_a_type.model == "interface":
            fp_to_interface_cable = needed_cables.pop(0)

            is_swapped = fp_to_interface_cable['termination_b'].device == open_element['device']
            intermediate_elements.append({
                **open_element,
                "attachment_b": fp_to_interface_cable['termination_b'] if is_swapped else fp_to_interface_cable['termination_a'],
            })

            intermediate_elements.append({
                "needed_cable": fp_to_interface_cable,
            })

            next_attachment_point = fp_to_interface_cable['termination_b'] if not is_swapped else fp_to_interface_cable['termination_a']

            open_element = {
                "attachment_a": next_attachment_point,
                "device": next_attachment_point.device,
            }
        elif self.termination_a_type.app_label == "dcim" and self.termination_a_type.model == "frontport":
            open_element["attachment_a"] = self.termination_a

        prepended_cable = None
        if self.termination_b_type.app_label == "dcim" and self.termination_b_type.model == "interface":
            prepended_cable = needed_cables.pop()


        for i, next_existing_cable in enumerate(cable_list):

            is_swapped = next_existing_cable.termination_b.device == open_element['device']

            open_element["attachment_b"] = next_existing_cable.termination_b if is_swapped else \
                next_existing_cable.termination_a

            intermediate_elements.append(open_element)

            intermediate_elements.append({
                "cable": next_existing_cable
            })

            next_attachment_point = next_existing_cable.termination_a if is_swapped else next_existing_cable.termination_b
            open_element = {
                "attachment_a": next_attachment_point,
                "device": next_attachment_point.device,
            }

            if i >= len(needed_cables):
                continue

            needed_cable = needed_cables[i]

            is_swapped = needed_cable['termination_b'].device == open_element['device']
            intermediate_elements.append({
                **open_element,
                "attachment_b": needed_cable['termination_b'] if is_swapped else needed_cable['termination_a'],
            })

            intermediate_elements.append({
                "needed_cable": needed_cable,
            })

            next_attachment_point = needed_cable['termination_b'] if not is_swapped else needed_cable['termination_a']

            open_element = {
                "attachment_a": next_attachment_point,
                "device": next_attachment_point.device,
            }

        if prepended_cable:

            is_swapped = prepended_cable['termination_b'].device == open_element['device']
            intermediate_elements.append({
                **open_element,
                "attachment_b": prepended_cable['termination_b'] if is_swapped else prepended_cable['termination_a'],
            })

            intermediate_elements.append({
                "needed_cable": prepended_cable,
            })

            next_attachment_point = prepended_cable['termination_b'] if not is_swapped else prepended_cable['termination_a']

            open_element = {
                "attachment_a": next_attachment_point,
                "device": next_attachment_point.device,
            }
        else:
            open_element["attachment_b"] = self.termination_b


        if open_element:
            intermediate_elements.append(open_element)

        return intermediate_elements

    def get_needed_cables(self):

        rearport_uuids = list()
        for c in self.cable_list:
            rearport_uuids.append(c.termination_a_id)
            rearport_uuids.append(c.termination_b_id)

        devices = Device.objects.filter(rearports__in=rearport_uuids)
        needed_cables = list()

        for rack in self.rack_list:

            if rack.id == self.termination_a.device.rack.id and self.termination_a_type.app_label == "dcim" and self.termination_a_type.model == "frontport":
                # We need no first cable to the FP, we ware already at the FP.
                continue
            elif rack.id == self.termination_b.device.rack.id and self.termination_b_type.app_label == "dcim" and self.termination_b_type.model == "frontport":
                # We need no first cable to the FP, we ware already at the FP.
                continue
            elif rack.id == self.termination_a.device.rack.id:
                # Start Rack
                pp_device = next(filter(lambda d: d.rack.id == rack.id, devices))
                fp_to_use = pp_device.frontports.filter(cable=None, rear_port__in=rearport_uuids).first()

                if not fp_to_use:
                    raise ValidationError(f"{pp_device} has no free ports.")

                needed_cables.append({
                    "termination_a_type": ContentType.objects.get_for_model(self.termination_a),
                    "termination_a_id": self.termination_a.id,
                    "termination_a": self.termination_a,
                    "termination_b_type": ContentType.objects.get_for_model(fp_to_use),
                    "termination_b_id": fp_to_use.id,
                    "termination_b": fp_to_use
                })
            elif rack.id == self.termination_b.device.rack.id:
                # Destination Rack
                pp_device = next(filter(lambda d: d.rack.id == rack.id, devices))
                fp_to_use = pp_device.frontports.filter(cable=None, rear_port__in=rearport_uuids).first()

                if not fp_to_use:
                    raise ValidationError(f"{pp_device} has no free ports.")

                needed_cables.append({
                    "termination_a_type": ContentType.objects.get_for_model(fp_to_use),
                    "termination_a_id": fp_to_use.id,
                    "termination_a": fp_to_use,
                    "termination_b_type": ContentType.objects.get_for_model(self.termination_b),
                    "termination_b_id": self.termination_b.id,
                    "termination_b": self.termination_b
                })
            else:
                # Hop Rack
                rack_devices = list(filter(lambda d: d.rack.id == rack.id, devices))
                fp1_to_use = FrontPort.objects.filter(cable=None, device=rack_devices[0],
                                                      rear_port__in=rearport_uuids).first()
                fp2_to_use = FrontPort.objects.filter(cable=None, device=rack_devices[1],
                                                      rear_port__in=rearport_uuids).first()

                if not fp1_to_use:
                    raise ValidationError(f"{rack_devices[0]} has no free ports.")
                if not fp2_to_use:
                    raise ValidationError(f"{rack_devices[1]} has no free ports.")

                needed_cables.append({
                    "termination_a_type": ContentType.objects.get_for_model(fp1_to_use),
                    "termination_a_id": fp1_to_use.id,
                    "termination_a": fp1_to_use,
                    "termination_b_type": ContentType.objects.get_for_model(fp2_to_use),
                    "termination_b_id": fp2_to_use.id,
                    "termination_b": fp2_to_use,
                })
        return needed_cables


class CableRouter:

    def __init__(self, termination_a, termination_a_type, termination_b, termination_b_type, media_type):
        self.termination_a = termination_a
        self.termination_b = termination_b

        self.allowed_cable_types = None
        if media_type == "fiber_sm":
            self.allowed_cable_types = [
                CableTypeChoices.TYPE_SMF,
                CableTypeChoices.TYPE_SMF_OS1,
                CableTypeChoices.TYPE_SMF_OS2,
            ]
        elif media_type == "fiber_mm":
            self.allowed_cable_types = [
                CableTypeChoices.TYPE_MMF,
                CableTypeChoices.TYPE_MMF_OM1,
                CableTypeChoices.TYPE_MMF_OM2,
                CableTypeChoices.TYPE_MMF_OM3,
                CableTypeChoices.TYPE_MMF_OM4,
            ]
        elif media_type == "copper":
            self.allowed_cable_types = [
                CableTypeChoices.TYPE_CAT3,
                CableTypeChoices.TYPE_CAT5,
                CableTypeChoices.TYPE_CAT5E,
                CableTypeChoices.TYPE_CAT6,
                CableTypeChoices.TYPE_CAT6A,
                CableTypeChoices.TYPE_CAT7,
                CableTypeChoices.TYPE_CAT7A,
                CableTypeChoices.TYPE_CAT8,
            ]

        excluded_cable_ids = list()
        if termination_a_type.app_label == "dcim" and termination_a_type.model == "frontport":
            rps_to_exclude = RearPort.objects.filter(device=self.termination_a.device)
            for rp in rps_to_exclude:
                if termination_a.rear_port != rp and rp.cable:
                    excluded_cable_ids.append(rp.cable.id)

        if termination_b_type.app_label == "dcim" and termination_b_type.model == "frontport":
            rps_to_exclude = RearPort.objects.filter(device=self.termination_b.device)
            for rp in rps_to_exclude:
                if termination_b.rear_port != rp and rp.cable:
                    excluded_cable_ids.append(rp.cable.id)

        possible_edges = Cable.objects.raw(
            '''
                SELECT c.*, d_a.rack_id as termination_a_rack_id, d_b.rack_id as termination_b_rack_id FROM dcim_cable c
                LEFT JOIN django_content_type ct_a ON ct_a.id = termination_a_type_id
                LEFT JOIN django_content_type ct_b ON ct_b.id = termination_b_type_id
                LEFT JOIN dcim_rearport rp_a ON rp_a.id = termination_a_id
                LEFT JOIN dcim_rearport rp_b ON rp_b.id = termination_b_id
                INNER JOIN dcim_frontport fp_a ON fp_a.rear_port_id = rp_a.id AND fp_a.cable_id IS NULL
                INNER JOIN dcim_frontport fp_b ON fp_b.rear_port_id = rp_b.id AND fp_b.cable_id IS NULL
                LEFT JOIN dcim_device d_a ON rp_a.device_id = d_a.id
                LEFT JOIN dcim_device d_b ON rp_b.device_id = d_b.id
                WHERE (
                    c.type = ANY(%s) AND
                    NOT c.id = ANY(%s) AND 
                    ct_a.app_label = 'dcim' AND ct_a.model = 'rearport' AND
                    ct_b.app_label = 'dcim' AND ct_b.model = 'rearport' AND
                    fp_a.cable_id IS NULL and fp_b.cable_id IS NULL
                )
            ''',
            [self.allowed_cable_types, excluded_cable_ids]
        )

        self.graph = Graph()
        for edge in possible_edges:
            self.graph.add_edge(edge.termination_a_rack_id, edge.termination_b_rack_id, edge.id)
            self.graph.add_edge(edge.termination_b_rack_id, edge.termination_a_rack_id, edge.id)

    def get_devices_for_hop_list(self, hop_list):
        devices = Device.objects.filter(id__in=hop_list)
        return list(map(lambda h: next(filter(lambda d: d.id == h, devices)), hop_list))

    def get_path(self):

        if self.termination_a.device.rack.id == self.termination_b.device.rack.id:
            raise ValidationError(
                f"{self.termination_a.device} and {self.termination_b.device} are located within the same rack")

        try:
            path = find_path(
                self.graph,
                self.termination_a.device.rack.id,
                self.termination_b.device.rack.id,
                cost_func=lambda a, b, c, d: 1
            )
        except NoPathError:
            raise ValidationError(
                f"No path available between {self.termination_a.device} and {self.termination_b.device}")

        return path.edges



import django_tables2 as tables
from django_tables2.utils import Accessor

from nautobot.dcim.tables.template_code import CABLE_TERMINATION_PARENT
from nautobot.utilities.tables import BaseTable, ColorColumn, ToggleColumn

from .models import CableTemplate, MeasurementLog

CABLE_LENGTH = """
{% if record.length %}{{ record.length }} {{ record.get_length_unit_display }}{% else %}&mdash;{% endif %}
"""


class CableTemplateTable(BaseTable):
    pk = ToggleColumn()
    name = tables.LinkColumn(
        'plugins:nautobot_cable_utils:cabletemplate_edit', args=[tables.A('pk')],
        verbose_name="Cable Number"
    )
    type = tables.Column()
    length = tables.TemplateColumn(
        template_code=CABLE_LENGTH,
    )
    color = ColorColumn()
    in_use = tables.TemplateColumn(
        template_code="<span>{% if record.cable %}✔{% else %}✘{% endif %}</span>",
        order_by="cable",
    )

    class Meta(BaseTable.Meta):
        model = CableTemplate
        fields = ('pk', 'name', 'label', 'owner', 'type', 'plug', 'length', 'color', 'in_use')
        default_columns = (
            'pk', 'name', 'label', 'owner', 'type', 'plug', 'length', 'color', 'in_use'
        )


class MeasurementLogTable(BaseTable):
    pk = ToggleColumn()
    link = tables.LinkColumn(
        'plugins:nautobot_cable_utils:measurement_log_edit', args=[tables.A('pk')],
    )
    cable = tables.LinkColumn('dcim:cable', args=[tables.A('pk')])

    class Meta(BaseTable.Meta):
        model = MeasurementLog
        fields = ('pk', 'link', 'cable')
        default_columns = ('pk', 'link', 'cable')


class NeededCableTable(tables.Table):
    class Meta:
        attrs = {
            "class": "table table-hover table-headings",
        }

    termination_a_parent = tables.TemplateColumn(
        template_code=CABLE_TERMINATION_PARENT,
        accessor=Accessor("termination_a"),
        orderable=False,
        verbose_name="Side A",
    )
    termination_a = tables.LinkColumn(
        accessor=Accessor("termination_a"),
        orderable=False,
        verbose_name="Termination A",
    )
    termination_b_parent = tables.TemplateColumn(
        template_code=CABLE_TERMINATION_PARENT,
        accessor=Accessor("termination_b"),
        orderable=False,
        verbose_name="Side B",
    )
    termination_b = tables.LinkColumn(
        accessor=Accessor("termination_b"),
        orderable=False,
        verbose_name="Termination B",
    )



from django.db import models

from nautobot.dcim.choices import *
from nautobot.dcim.models import Cable
from nautobot.utilities.fields import ColorField
from nautobot.utilities.querysets import RestrictedQuerySet
from nautobot.core.models import BaseModel


class CablePlug(BaseModel):
    name = models.CharField(
        max_length=50,
        unique=True,
    )

    def __str__(self):
        return self.name


class CableTemplate(BaseModel):
    name = models.CharField(
        max_length=100,
        unique=True,
    )
    label = models.CharField(
        max_length=100,
        blank=True
    )
    type = models.CharField(
        max_length=50,
        choices=CableTypeChoices,
        blank=True
    )
    plug = models.ForeignKey(
        to=CablePlug,
        on_delete=models.CASCADE,
        related_name="cable_templates",
        blank=True,
        null=True,
    )
    color = ColorField(
        blank=True
    )
    length = models.PositiveSmallIntegerField(
        blank=True,
        null=True
    )
    length_unit = models.CharField(
        max_length=50,
        choices=CableLengthUnitChoices,
        blank=True,
    )
    cable = models.OneToOneField(
        Cable,
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
    )
    owner = models.ForeignKey(
        to="tenancy.Tenant",
        on_delete=models.CASCADE,
        related_name="cable_templates",
        blank=True,
        null=True,
    )

    objects = RestrictedQuerySet.as_manager()

    csv_headers = [
        'name', 'owner', 'type', 'plug', 'label', 'color', 'length', 'length_unit',
    ]

    def __str__(self):
        return self.name


class MeasurementLog(BaseModel):
    link = models.URLField(
        blank=True,
        null=True
    )
    cable = models.OneToOneField(
        Cable,
        on_delete=models.CASCADE,
    )
